
/* GNU djinn - UNIX daemon wrapper for Java programs
   Copyright (C) 1998 PING Software Group and Mark Lindner

   This file is part of GNU djinn
   
   GNU djinn is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License, or (at your option) any later version.

   GNU djinn is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU djinn; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   The author may be contacted at:
 
   frenzy@ix.netcom.com
*/

/* --- System Headers --- */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <stddef.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#include <signal.h>

/* --- Local Headers --- */

#include "vector.h"
#include "getopt.h"

/* --- Macros --- */

#define ENV_JAVA "JAVA"
#define ENV_JAVA_ARGS "JAVA_ARGS"
#define DEVNULL "/dev/null"
#define VM_NAME "javaXXXX"

#define HELP_TEXT "Usage: %s [-p classpath] [-j jvm] [-a vmargs]\n\t[-o file] [--version --help --verbose] class\n"
#define VERSION_TEXT "djinn v1.1 - Mark Lindner - 12/10/98\n"

#define HELP_OPTION 0
#define VERSION_OPTION 1
#define VERBOSE_OPTION 2

#ifndef HAVE_STRDUP
char *djinn_strdup(const char *s);
#define strdup djinn-strdup
#endif /* HAVE_STRDUP */

static const struct option options[] = { { "help", 0, NULL, 0  },
                                         { "version", 0, NULL, 0 },
                                         { "verbose", 0, NULL, 0 },
                                         { NULL, 0, NULL, 0 } };

/* --- Functions --- */

int main(int argc, char **argv)
  {
  char *javavm = NULL, *classpath = NULL, *class = NULL, *logfile = NULL,
    *vmargs = NULL;
  int c, option_index, fd = -1, vflag = 0, errflag = 0;
  pid_t pid;

  void djinn_exec(char *, char *, char *, char *);

/* parse the command line */

  while((c = getopt_long(argc, argv, "a:p:j:o:", options, &option_index))
        != EOF)
    {
    switch(c)
      {
      case 'a':
        vmargs = optarg;
        break;
      case 'p':
        classpath = optarg;
        break;
      case 'j':
        javavm = optarg;
        break;
      case 'o':
        logfile = optarg;
        break;

      case 0:
        {
        switch(option_index)
          { 
          case HELP_OPTION:
            fprintf(stderr, HELP_TEXT, *argv);
            exit(EXIT_SUCCESS);
            break;

          case VERSION_OPTION:
            fprintf(stderr, VERSION_TEXT);
            exit(EXIT_SUCCESS);
            break;

          case VERBOSE_OPTION:
            vflag = 1;
            break;
            
          default:
            errflag = 1;
            break;
          }
        break;
        }

      default:
        errflag = 1;
        break;
      }
    }

  /* catch illegal option errors */

  if(errflag)
    {
    fprintf(stderr, HELP_TEXT, *argv);
    exit(EXIT_FAILURE);
    }
  
  /* parse out class file */

  if(argc <= optind)
    {
    fprintf(stderr, "%s: No Java class file specified.\n", *argv);
    exit(EXIT_FAILURE);
    }
  else
    {
    class = argv[optind];
    if(vflag)
      fprintf(stderr, "%s: Executing class file: %s\n", *argv, class);
    }

  /* check for optional java args; first check command line, and if the
     -a switch wasn't specified, resort to the environment variable
     JAVA_ARGS
  */

  if(!vmargs)
    vmargs = getenv(ENV_JAVA_ARGS);
  
  /* find the Java VM; first check the command line, and if the -j switch
     wasn't specified, resort to the environment variable JAVA
  */

  if(!javavm)
    javavm = getenv(ENV_JAVA);

  /* see if VM is valid */

  if(javavm)
    if(access(javavm, X_OK))
      {
      fprintf(stderr, "%s: Cannot execute VM: %s\n", *argv, javavm);
      exit(EXIT_FAILURE);
      }

  if(vflag)
    fprintf(stderr, "%s: Using VM: %s\n", *argv, (javavm ? javavm : VM_NAME));

  /* open logfile, if one specified */

  if(!logfile) logfile = strdup(DEVNULL);

  if((fd = open(logfile, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR)) < 0)
    {
    fprintf(stderr, "%s: Unable to open logfile: %s\n", *argv, logfile);
    exit(EXIT_FAILURE);
    }
  else
    {
    if(vflag)
      fprintf(stderr, "%s: Using logfile: %s\n", *argv, logfile);
    }

/* other miscellany */
      
  signal(SIGHUP, SIG_IGN);

/* fork subprocess and run VM */
      
  pid = fork();
  switch(pid)
    {
    case 0:
      {
      /* child */
        
      (void)setsid();      
      dup2(fd, STDOUT_FILENO);
      dup2(fd, STDERR_FILENO);
      close(STDIN_FILENO);
      
      djinn_exec(javavm, vmargs, classpath, class);
      exit(EXIT_FAILURE);
      }
      
    case -1:
      {
      /* fork error */

      fprintf(stderr, "%s: Unable to fork subprocess.\n", *argv);
      exit(EXIT_FAILURE);
      break;
      }
      
    default:
      {
      /* parent */
      
      if(vflag)
        fprintf(stderr, "%s: Daemon process started (pid: %i)\n", *argv,
                (int)pid);
      break;
      }
    }

  exit(EXIT_SUCCESS);
  }
  
void djinn_exec(char *javavm, char *vmargs, char *classpath, char *class)
  {
  char *p, **args;
  vec_t *argvec;

  /* construct argument vector */
  
  argvec = vec_start(20);
  
  // java [-classpath CLASSPATH] [OTHER ARGS] CLASS
  
  vec_store(argvec, VM_NAME);
  if(classpath != NULL)
    {
    // add classpath switch if classpath specified
     
    vec_store(argvec, "-classpath");
    vec_store(argvec, classpath);
    }
    
  // add any other args to the list
     
  if(vmargs != NULL)
    {
    for(p = strtok((char *)vmargs, " "); p; p = strtok(NULL, " "))
      vec_store(argvec, p);
    }

  vec_store(argvec, class);

  fprintf(stderr, "Invoking: ");
  vec_print(stderr, argvec);
  fputc('\n', stderr);
  
  args = vec_end(argvec, NULL);
  
  if(javavm == NULL)
    execvp(VM_NAME, args);
  else
    execvp(javavm, args);

  fprintf(stderr, "Unable to start VM: ");
  perror(NULL);
  }

char *djinn_strdup(const char *s)
  {
  char *t = (char *)malloc(strlen(s) + 1);
  strcpy(t, s);
  return(t);
  }

/* --- end of source --- */
  
  
